import {
  Body,
  Controller,
  ForbiddenException,
  Get,
  Param,
  Post,
  Req,
  Sse,
  UseGuards,
} from '@nestjs/common';
import { AuthGuard } from 'src/Auth/auth/auth.guard';
import { UserGuard } from '../user/user.guard';
import { UserRequest } from 'src/utils/UserRequest';
import { PartyService } from './party.service';
import { UserService } from '../user/user.service';
import { SseService } from './sse.service';
import { Subject } from 'rxjs';
import { RunService } from '../run/run.service';
import { PartyStatus } from '@prisma/client';

@Controller('party')
export class PartyController {
  constructor(
    private readonly partyService: PartyService,
    private readonly userService: UserService,
    private readonly sseService: SseService,
    private readonly runService: RunService,
  ) {}

  @Sse('join/:party_id')
  @UseGuards(AuthGuard, UserGuard)
  async join(@Param('party_id') party_id: string, @Req() req: UserRequest) {
    const client = new Subject<any>();

    const userId = req.userId;
    console.log('Joining party', party_id, 'as user', userId);

    // verify if the party exists and is not closed
    try {
      const isPartyOpen = await this.partyService.isPartyOpen(party_id);

      if (!isPartyOpen) {
        throw new ForbiddenException('Party is closed');
      }
    } catch {
      throw new ForbiddenException('Party does not exist');
    }

    // join the party

    const party = await this.partyService.joinParty({
      userId,
      partyId: party_id,
    });

    // get their user data
    const users = await this.userService.findManyByPartyId(party.id);
    this.sseService.addClient(party_id, client);

    console.log(users);
    // send to the other users in the party that a new user has joined, using SSE
    setTimeout(() => {
      this.sseService.sendEvent(party_id, { type: 'userJoined', data: users });
    }, 1); // Délai de 100 ms, ajustez si nécessaire

    return client.asObservable();
  }

  @Post()
  @UseGuards(AuthGuard, UserGuard)
  async createOne(@Req() req: UserRequest, @Body() body: any) {
    const { quizId } = body;

    const userId = req.userId;

    console.log('Creating party for user', userId);

    // create the party
    const party = await this.partyService.createParty({
      hostId: userId,
      quizId,
    });

    return party;
  }

  @Get('generateRun')
  async generateRun() {
    return 'Unused function';
  }

  @Get('start')
  @UseGuards(AuthGuard, UserGuard)
  async startGame(@Req() req: UserRequest) {
    const userId = req.userId;

    const user = await this.userService.findOneWithPartyInfos(userId);

    if (user.isHost === false) {
      throw new ForbiddenException('User is not the host');
    }

    try {
      await this.partyService.setPartyStatus(
        user.Party.id,
        PartyStatus.in_game,
      );
    } catch {
      throw new ForbiddenException('Party does not exist');
    }

    const players = await this.userService.findManyByPartyId(user.Party.id);

    for (const player of players) {
      await this.partyService.createRun(player.id, user.Party.id);
    }

    const run = await this.runService.getRunByPartyAndHost(userId);

    this.sseService.sendEvent(user.Party.id, {
      type: 'gameStarted',
      data: {
        questionCount: run.quiz.questionCount,
      },
    });

    const question = await this.partyService.getQuestion({
      id: run.id,
    });

    this.sseService.sendEvent(user.Party.id, {
      type: 'nextQuestion',
      data: question,
    });
  }

  @Get('question')
  @UseGuards(AuthGuard, UserGuard)
  async getQuestion(@Req() req: UserRequest) {
    const userId = req.userId;

    const run = await this.runService.getRunByPartyAndHost(userId);

    const question = await this.partyService.getQuestion({
      id: run.id,
    });

    return { question };
  }

  @Get('nextQuestion')
  @UseGuards(AuthGuard, UserGuard)
  async nextQuestion(@Req() req: UserRequest) {
    const userId = req.userId;

    const user = await this.userService.findOneWithPartyInfos(userId);

    const run = await this.runService.getRunByPartyAndHost(userId);

    try {
      const question = await this.partyService.getQuestion({
        id: run.id,
      });

      this.sseService.sendEvent(user.partyId, {
        type: 'nextQuestion',
        data: question,
      });
    } catch {
      // if there are no question, send the end of the game
      const scores = await this.partyService.getScoreboard(user.partyId);

      this.sseService.sendEvent(user.partyId, {
        type: 'gameEnded',
        data: scores,
      });
    }
  }

  @Get('leave/:party_id')
  @UseGuards(AuthGuard, UserGuard)
  async leave(@Param('id') party_id: string, @Req() req: UserRequest) {
    const userId = req.userId;
    console.log('Leaving party', party_id, 'as user', userId);

    // verify if the party exists and is not closed
    const isPartyOpen = await this.partyService.isPartyOpen(party_id);

    if (!isPartyOpen) {
      throw new ForbiddenException('Party is closed');
    }

    // leave the party

    const party = await this.partyService.leaveParty({
      userId,
    });

    // get their user data
    const user = await this.userService.findOne(userId);

    // send to the other users in the party that a user has left, using SSE
    this.sseService.sendEvent(party_id, { type: 'userLeft', user });

    return party;
  }

  @Get('/score')
  @UseGuards(AuthGuard, UserGuard)
  async getScore(@Req() req: UserRequest) {
    const userId = req.userId;

    const user = await this.userService.findOneWithPartyInfos(userId);

    const run = await this.runService.getRunByPartyAndHost(user.id);

    return { score: run.score };
  }

  @Get('/scoreboard')
  @UseGuards(AuthGuard, UserGuard)
  async getScoreboard(@Req() req: UserRequest) {
    const userId = req.userId;

    const user = await this.userService.findOneWithPartyInfos(userId);

    const users = await this.userService.findManyByPartyId(user.Party.id);

    const scores = await this.partyService.getScoreboard(user.Party.id);

    return { users, scores };
  }

  @Post('answer')
  @UseGuards(AuthGuard, UserGuard)
  async answer(@Req() req: UserRequest, @Body() body: any) {
    const { questionId, answerId } = body;
    const user = await this.userService.findOneWithPartyInfos(req.userId);

    const run = await this.runService.getRunByPartyAndHost(user.id);

    await this.partyService.answer({
      runId: run.id,
      questionId,
      answerId,
    });
  }

  @Get('revealAnswer')
  @UseGuards(AuthGuard, UserGuard)
  async revealAnswer(@Req() req: UserRequest) {
    const user = await this.userService.findOneWithPartyInfos(req.userId);

    const run = await this.runService.getRunByPartyAndHost(user.id);
    console.log(run);

    const questionId = await this.runService.getAnsweredQuestionId(run.id);
    const question = await this.runService.getAnswers('', questionId);

    console.log(question);
    this.sseService.sendEvent(user.Party.id, {
      type: 'answerRevealed',
      data: question,
    });
  }
}
