import { PrismaService } from 'src/infra/prisma/prisma-service';
import { PartyRepository } from './party-repository';
import { PartyStatus } from '@prisma/client';
import { Injectable } from '@nestjs/common';
import { RunService } from 'src/domain/run/run.service';

@Injectable()
export class PartyRepositoryOnDb implements PartyRepository {
  constructor(
    private readonly prisma: PrismaService,
    private readonly runService: RunService,
  ) {}

  async findParty(partyId: string) {
    return await this.prisma.party.findUnique({
      where: {
        id: partyId,
      },
    });
  }

  async createParty(infos: {
    hostId: number;
    quizId: string;
    partyId: string;
  }) {
    const { hostId, quizId, partyId } = infos;
    const party = await this.prisma.party.create({
      data: {
        id: partyId,
        name: 'New party',
        quizId,
      },
    });

    // make the user join
    await this.prisma.quizUser.update({
      where: {
        id: hostId,
      },
      data: {
        partyId: partyId,
        isHost: true,
      },
    });

    return party;
  }

  async joinParty(infos: { userId: number; partyId: string }) {
    const { userId, partyId } = infos;

    const party = await this.prisma.party.update({
      where: {
        id: partyId,
      },
      data: {
        players: {
          connect: {
            id: userId,
          },
        },
      },
      select: {
        players: true,
      },
    });

    if (party.players.length === 1) {
      await this.prisma.quizUser.update({
        where: {
          id: userId,
        },
        data: {
          isHost: true,
        },
      });
    } else if (party.players.length > 1) {
      await this.prisma.quizUser.update({
        where: {
          id: userId,
        },
        data: {
          partyId,
          isHost: false,
        },
      });
    }

    // return party without the players
    return await this.prisma.party.findUnique({
      where: {
        id: partyId,
      },
    });
  }

  async leaveParty(userId: number): Promise<void> {
    await this.prisma.quizUser.update({
      where: {
        id: userId,
      },
      data: {
        partyId: null,
        isHost: false,
      },
    });
  }

  async setPartyStatus(partyId: string, status: PartyStatus): Promise<void> {
    await this.prisma.party.update({
      where: {
        id: partyId,
      },
      data: {
        status,
      },
    });
  }

  async getQuizId(partyId: string): Promise<string> {
    const party = await this.prisma.party.findUnique({
      where: {
        id: partyId,
      },
    });

    return party.quizId;
  }

  async answer({
    runId,
    questionId,
    answerId,
  }: {
    runId: string;
    questionId: number;
    answerId: number;
  }) {
    const question = await this.prisma.question.findUnique({
      where: {
        id: questionId,
      },
      include: {
        answers: true,
      },
    });

    if (!question) {
      throw new Error('Question not found');
    }

    const isCorrect = question.answers.some(
      (a) => a.isCorrect && a.id === answerId,
    );

    await this.prisma.run.update({
      where: {
        id: runId,
      },
      data: {
        questionIndex: {
          increment: 1,
        },
      },
    });
    if (isCorrect) {
      await this.runService.incrementScore(runId);
    }
  }

  async getScores(partyId: string) {
    const runs = await this.prisma.run.findMany({
      where: {
        partyId,
      },
      select: {
        score: true,
        QuizUser: true,
      },
      orderBy: {
        score: 'desc',
      },
    });

    return runs.map((r) => ({
      id: r.QuizUser.id,
      username: r.QuizUser.username,
      score: r.score,
    }));
  }
}
