import { forwardRef, Inject, Injectable } from '@nestjs/common';
import { CreateQuizDto } from './dto/create-quiz.dto';
import { UpdateQuizDto } from './dto/update-quiz.dto';
import { QuizRepository } from './store/quiz-repository';
import * as he from 'he';
import { Quiz, Question, Category, Difficulty, Answer } from '@prisma/client';
import { CreateCommunityQuizDto } from './dto/community-quiz.dto';
import { UpdateQuestionDto } from '../question/dto/update-question.dto';
import { RunService } from '../run/run.service';

type FullQuiz = Quiz & {
  category?: Category | null;
  difficulty?: Difficulty | null;
  questions: FullQuestion[];
};

type FullQuestion = Question & {
  answers?: Answer[] | null;
  category?: Category | null;
  difficulty?: Difficulty | null;
};

type AnswerWithoutIsCorrect = Omit<Answer, 'isCorrect'>;

type FullQuestionWithoutIsCorrect = Omit<FullQuestion, 'answers'> & {
  answers?: AnswerWithoutIsCorrect[];
};

type FullQuizWithoutIsCorrect = Omit<FullQuiz, 'questions'> & {
  questions: FullQuestionWithoutIsCorrect[];
};

@Injectable()
export class QuizService {
  constructor(
    private readonly quizRepository: QuizRepository,
    @Inject(forwardRef(() => RunService))
    private readonly runService: RunService,
  ) {}

  async getRemainingQuestions(id: string) {
    const quiz = await this.quizRepository.findByIdWithoutAnswers(id);
    if (!quiz) {
      throw new Error('Quiz not found');
    }

    const filteredQuiz = {
      ...quiz,
    };

    return filteredQuiz;
  }

  // genere un quiz avec des questions de trivia et lance une run dessus
  async importQuestions(
    amount: number,
    category: number,
    difficulty: string,
    user: any,
  ) {
    let url = 'https://opentdb.com/api.php?';
    if (amount) url += `amount=${amount}`;
    if (category) url += `&category=${category + 8}`;
    if (difficulty) url += `&difficulty=${difficulty}`;

    const response = await fetch(url);
    const triviaData = await response.json();

    // Transformer les données reçues en quiz
    const quiz = this.transformTriviaData(triviaData.results);
    quiz.id = await this.generateUniqueId();

    if (user) {
      quiz.authorId = user;
    }

    // Sauvegarder le quiz original
    await this.quizRepository.save(quiz);

    return await this.startRun(quiz.id, user);
  }

  // Omet la propriété `isCorrect` des réponses
  async omitCorrectAnswer(quiz: FullQuiz): Promise<FullQuizWithoutIsCorrect> {
    const questions = quiz.questions.map((question) => {
      const answers = question.answers.map((answer) => ({
        id: answer.id,
        questionId: answer.questionId,
        text: answer.text, // Omet la propriété `isCorrect`
        fileType: answer.fileType,
      }));

      return {
        ...question,
        answers, // Remplace les réponses sans `isCorrect`
      };
    });

    const omitQuiz: FullQuizWithoutIsCorrect = {
      ...quiz,
      questions, // Remplace les questions avec les nouvelles sans `isCorrect`
    };

    return omitQuiz;
  }

  // lance une run sur un quiz donné
  async startRun(id: string, user: any, roomId?: string) {
    const quiz = await this.quizRepository.findById(id);
    const runQuizId = await this.generateUniqueId();
    const runId = await this.generateUniqueId();

    if (!quiz) {
      throw new Error('Quiz not found');
    }
    await this.duplicateQuiz(id, runQuizId);

    const runQuiz = await this.quizRepository.findById(runQuizId);

    if (user) {
      const run = await this.runService.create(
        runId,
        runQuiz.id,
        quiz.id,
        user,
        roomId,
      );
      return {
        ...run,
        QuestionCount: quiz.questionCount,
      };
    } else {
      const run = await this.runService.createWithoutLog(
        runId,
        runQuiz.id,
        quiz.id,
        roomId,
      );
      return {
        ...run,
        QuestionCount: quiz.questionCount,
      };
    }
  }

  // importe uniquement les questions trivia
  async importQuestionsForCommunity(
    amount: number,
    category: number,
    difficulty: string,
  ) {
    let url = 'https://opentdb.com/api.php?';
    if (amount) url += `amount=${amount}`;
    if (category) url += `&category=${category + 8}`;
    if (difficulty) url += `&difficulty=${difficulty}`;

    const response = await fetch(url);
    const triviaData = await response.json();

    const triviaQuestions = this.transformTriviaData(triviaData.results);
    return triviaQuestions;
  }

  // transforme les données trivia en quiz
  private transformTriviaData(triviaData: any[]): FullQuiz {
    const questions = triviaData.map((data, index) => {
      let answers: Answer[];

      if (data.type === 'boolean') {
        answers = [
          {
            id: null,
            questionId: null,
            text: 'True',
            isCorrect: data.correct_answer === 'True',
            fileType: 'text',
          },
          {
            id: null,
            questionId: null,
            text: 'False',
            isCorrect: data.correct_answer === 'False',
            fileType: 'text',
          },
        ];
      } else {
        answers = [
          {
            id: null,
            questionId: null,
            text: he.decode(data.correct_answer),
            isCorrect: true,
            fileType: 'text',
          },
          ...data.incorrect_answers.map(
            (answer) =>
              ({
                id: null,
                questionId: null,
                text: he.decode(answer),
                isCorrect: false,
              }) as Answer,
          ),
        ];
      }

      return {
        id: -1,
        text: he.decode(data.question),
        type: data.type,
        categoryId: -1,
        Category: {
          id: -1,
          name: he.decode(data.category),
        },
        quizId: '',
        order: index + 1,
        answers: answers,
      };
    });

    const category: Category = { id: 0, name: triviaData[0].category };
    const difficulty: Difficulty = { id: 0, name: triviaData[0].difficulty };

    const quiz: FullQuiz = {
      id: '',
      name: 'New Quiz',
      categoryId: 0,
      description: 'Description of the quiz',
      isCommunity: false,
      questionCount: questions.length,
      difficultyId: 0,
      authorId: 0,
      category: {
        id: 0,
        name: category.name,
      },
      difficulty: {
        id: 0,
        name: difficulty.name,
      },
      questions: questions,
      createdAt: new Date(),
      updatedAt: new Date(),
    };

    return quiz;
  }

  // Crée un quiz communautaire
  async createCommunityQuiz(
    createCommunityQuizDto: CreateCommunityQuizDto,
    user: any,
  ) {
    const { name, description, difficulty, questions } = createCommunityQuizDto;

    const difficultyEntity: Difficulty = { id: null, name: difficulty };

    // Transformer les questions reçues
    const transformedQuestions = questions.map((question, index) => {
      const answers = question.answers.map((answer) => {
        return {
          id: null,
          questionId: null,
          text: answer.text,
          isCorrect: answer.isCorrect ?? false,
          fileType: answer.fileType ?? 'text',
        };
      });

      return {
        id: null, // ID null avant sauvegarde
        text: question.text,
        type: question.isMultipleChoice ? 'multiple' : 'boolean',
        answers: answers,
        order: index + 1, // Ordre
        Category: null, // Catégorie null car non spécifiée
        categoryId: -1, // Placeholder value for categoryId
        quizId: '', // Placeholder value for quizId
      };
    });

    const questionCount = transformedQuestions.length;
    const authorId = user;

    if (!name) {
      throw new Error('Name is required');
    }
    if (!description) {
      throw new Error('Description is required');
    }

    const quiz: FullQuiz = {
      id: await this.generateUniqueId(),
      name,
      description,
      isCommunity: true,
      questionCount,
      categoryId: -1, // Placeholder value for categoryId
      difficultyId: difficultyEntity.id,
      category: null,
      difficulty: difficultyEntity,
      questions: transformedQuestions,
      authorId,
      createdAt: new Date(),
      updatedAt: new Date(),
    };

    await this.quizRepository.save(quiz);

    return quiz;
  }

  async updateQuestion(
    quizId: string,
    questionId: number,
    updateQuestionDto: UpdateQuestionDto,
  ) {
    const { text, isMultipleChoice, answers } = updateQuestionDto;

    // Transformer les réponses pour correspondre au format attendu
    const transformedAnswers = answers.map((answer) => ({
      text: answer.text,
      isCorrect: answer.isCorrect ?? false, // Défaut : incorrect si non spécifié
    }));

    // Appeler le repository pour effectuer la mise à jour
    return this.quizRepository.updateQuestions(quizId, questionId, {
      text,
      type: isMultipleChoice ? 'multiple' : 'boolean',
      answers: transformedAnswers,
    });
  }

  async generateUniqueId(): Promise<string> {
    const generateCode = () => {
      const characters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';
      let code = '';
      for (let i = 0; i < 6; i++) {
        code += characters.charAt(
          Math.floor(Math.random() * characters.length),
        );
      }
      return code;
    };

    let id;
    let isUnique = false;

    while (!isUnique) {
      id = generateCode();
      const existingQuiz = await this.quizRepository.findById(id);
      if (!existingQuiz) {
        isUnique = true;
      }
    }

    return id;
  }

  async getCategories() {
    await this.importCategories();
    return this.quizRepository.getCategories();
  }

  //utiliser cette fonction quand on veut update toutes les catégories
  async importCategories() {
    const url = 'https://opentdb.com/api_category.php';
    const response = await fetch(url);
    const triviaData = await response.json();

    // Parcourt chaque catégorie et les sauvegarde
    for (const category of triviaData.trivia_categories) {
      // Utilise l'id et le name si nécessaire
      await this.quizRepository.saveCategory(he.decode(category.name));
    }
  }

  async duplicateQuiz(id: string, runQuizId: string): Promise<void> {
    return this.quizRepository.duplicateQuiz(id, runQuizId);
  }

  getAnswers(codeQuiz: string, questionID: number) {
    return this.quizRepository.getAnswers(codeQuiz, questionID);
  }

  getStats(id: string) {
    return this.quizRepository.getStats(id);
  }

  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  create(createQuizDto: CreateQuizDto) {
    return 'This action adds a new quiz';
  }

  findAll() {
    return this.quizRepository.findAll();
  }

  findOne(id: string) {
    return this.quizRepository.findById(id);
  }

  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  update(id: number, updateQuizDto: UpdateQuizDto) {
    return `This action updates a #${id} quiz`;
  }

  remove(id: number) {
    return `This action removes a #${id} quiz`;
  }

  async getPaginatedQuizzes(skip: number, take: number, search: string) {
    return search === ''
      ? await this.quizRepository.getPaginated(skip, take)
      : await this.quizRepository.getPaginatedBySearch(skip, take, search);
  }
  async info(quizId: string) {
    return this.quizRepository.getInfo(quizId);
  }

  getUserCustomQuiz(userId: string) {
    return this.quizRepository.getUserCustomQuiz(userId);
  }
}
