import { Injectable } from '@nestjs/common';
import { PrismaService } from '../../../infra/prisma/prisma-service';
import { QuizRepository } from './quiz-repository';
import { Quiz, Answer, Category, Difficulty, Question } from '@prisma/client';

type FullQuestion = Question & {
  answers: Omit<Answer, 'isCorrect'>[] | Answer[];
  category?: Category | null;
  difficulty?: Difficulty | null;
};

type FullQuiz = Quiz & {
  category?: Category | null;
  difficulty?: Difficulty | null;
  questions: FullQuestion[];
};

@Injectable()
export class QuizRepositoryOnDb implements QuizRepository {
  constructor(private readonly prisma: PrismaService) {}

  async save(quiz: FullQuiz): Promise<void> {
    await this.prisma.quiz.create({
      data: {
        id: quiz.id,
        name: quiz.name,
        description: quiz.description,
        questionCount: quiz.questionCount,
        category: quiz.category?.name
          ? {
              connectOrCreate: {
                where: { name: quiz.category.name },
                create: { name: quiz.category.name },
              },
            }
          : undefined, // Si aucune catégorie, ne pas inclure
        difficulty: quiz.difficulty?.name
          ? {
              connectOrCreate: {
                where: { name: quiz.difficulty.name },
                create: { name: quiz.difficulty.name },
              },
            }
          : undefined, // Si aucune difficulté, ne pas inclure
        questions: {
          create: quiz.questions.map((question, index) => ({
            text: question.text,
            type: question.type,
            order: index,
            category: quiz.category?.name
              ? {
                  connectOrCreate: {
                    where: { name: quiz.category.name },
                    create: { name: quiz.category.name },
                  },
                }
              : undefined,
            answers: {
              create: question.answers.map((answer) => ({
                text: answer.text,
                isCorrect: answer.isCorrect || false,
                fileType: answer.fileType,
              })),
            },
          })),
        },
        isCommunity: quiz.isCommunity,
        author: quiz.authorId ? { connect: { id: quiz.authorId } } : undefined, // Si pas d'auteur, ignorer
      },
    });
  }
  async duplicateQuiz(originalQuizId: string, newid: string): Promise<void> {
    // Récupérer le quiz original avec toutes ses relations
    const originalQuiz = await this.findById(originalQuizId);
    if (!originalQuiz) {
      throw new Error('Original quiz not found');
    }

    await this.prisma.quiz.create({
      data: {
        id: newid,
        name: originalQuiz.name,
        description: originalQuiz.description,
        isCommunity: false,
        questionCount: originalQuiz.questionCount,
        category: originalQuiz.categoryId
          ? {
              connect: { id: originalQuiz.categoryId },
            }
          : undefined, // Si pas de categoryId, ignorez cette propriété
        difficulty: originalQuiz.difficultyId
          ? {
              connect: { id: originalQuiz.difficultyId },
            }
          : undefined, // Si pas de difficultyId, ignorez cette propriété
        author: originalQuiz.authorId
          ? {
              connect: { id: originalQuiz.authorId },
            }
          : undefined, // Si pas d'auteur, ignorez cette propriété
        questions: {
          create: originalQuiz.questions.map((question) => ({
            text: question.text,
            type: question.type,
            order: question.order,
            category: question.categoryId
              ? {
                  connect: { id: question.categoryId },
                }
              : undefined, // Si pas de categoryId, ignorez cette propriété
            answers: {
              create: question.answers.map((answer) => ({
                text: answer.text,
                isCorrect: answer.isCorrect, // Sauvegarde initiale avec `isCorrect`
                fileType: answer.fileType,
              })),
            },
          })),
        },
      },
      include: {
        questions: {
          include: {
            answers: true,
            category: true,
          },
        },
        category: true,
        difficulty: true,
      },
    });
  }

  async findAll(): Promise<Quiz[]> {
    const allQuizzes = await this.prisma.quiz.findMany({
      include: {
        category: true,
        difficulty: true,
        questions: {
          include: {
            answers: true,
            category: true,
          },
        },
        author: true,
      },
    });

    return allQuizzes;
  }

  async findById(id: string): Promise<FullQuiz | null> {
    const quiz = await this.prisma.quiz.findUnique({
      where: { id },
      include: {
        category: true,
        difficulty: true,
        questions: {
          include: {
            answers: {
              select: {
                id: true,
                text: true,
                isCorrect: true,
                questionId: true,
                fileType: true,
              },
            },
            category: true,
          },
        },
      },
    });

    return quiz;
  }

  async findByIdWithoutAnswers(id: string): Promise<FullQuiz | null> {
    const quiz = await this.prisma.quiz.findUnique({
      where: { id },
      include: {
        category: true,
        difficulty: true,
        questions: {
          include: {
            answers: {
              omit: {
                isCorrect: true,
              },
            },
            category: true,
          },
        },
      },
    });

    return quiz;
  }

  async getStats(id: string): Promise<any> {
    // nombre de joueurs ayant joué au quiz (via la table run)
    // poucentage de bonne réponse
    try {
      const stats = await this.prisma.$queryRaw`
        SELECT (SELECT COUNT(*) FROM Run WHERE quizId = ${id}) as totalRuns,
        SELECT (SELECT COUNT(*) FROM AnswerRun WHERE isCorrect = true) as correctAnswers,
        SELECT (SELECT COUNT(*) FROM AnswerRun WHERE isCorrect = false) as incorrectAnswers
      `;

      return {
        totalRuns: Number(stats[0].totalRuns),
        correctAnswers: Number(stats[0].correctAnswers),
        incorrectAnswers: Number(stats[0].incorrectAnswers),
        totalAnswers: Number(
          stats[0].correctAnswers + stats[0].incorrectAnswers,
        ),
      };
    } catch {
      throw new Error('Failed to get stats');
    }
  }

  async getAnswers(
    id: string,
    questionId: number,
  ): Promise<Omit<Answer, 'isCorrect'>[]> {
    const answers = await this.prisma.answer.findMany({
      where: { questionId },
      omit: {
        isCorrect: false,
      },
    });

    return answers;
  }

  async updateQuestions(
    quizId: string,
    questionId: number,
    data: {
      text: string;
      type: string;
      answers: { text: string; isCorrect: boolean }[];
    },
  ): Promise<void> {
    // Supprimer les anciennes réponses liées à la question
    await this.prisma.answer.deleteMany({
      where: { questionId: Number(questionId) }, // Conversion explicite
    });

    // Mettre à jour la question et recréer les réponses
    await this.prisma.question.update({
      where: { id: Number(questionId) }, // Conversion explicite
      data: {
        text: data.text,
        type: data.type,
        answers: {
          create: data.answers.map((answer) => ({
            text: answer.text,
            isCorrect: answer.isCorrect,
          })),
        },
      },
    });
  }

  async getPaginated(
    skip: number,
    take: number,
  ): Promise<{ quizzes: Quiz[]; total: number }> {
    // Compter le nombre total de quiz
    const total = await this.prisma.quiz.count({
      where: {
        isCommunity: true,
      },
    });

    // Récupérer les quiz paginés
    const quizzes = await this.prisma.quiz.findMany({
      where: {
        isCommunity: true,
      },
      include: {
        category: true,
        difficulty: true,
        questions: {
          include: {
            answers: true,
          },
        },
        author: true,
      },
      skip,
      take,
    });

    // Retourner les quiz et le total
    return { quizzes, total };
  }

  async getPaginatedBySearch(
    skip: number,
    take: number,
    search: string,
  ): Promise<{ quizzes: Quiz[]; total: number }> {
    // Compter le nombre total de quiz correspondant aux critères
    const total = await this.prisma.quiz.count({
      where: {
        isCommunity: true,
        OR: [
          { name: { contains: search } },
          { description: { contains: search } },
          { difficulty: { name: { contains: search } } },
          { author: { username: { contains: search } } },
        ],
      },
    });

    // Récupérer les quiz paginés
    const quizzes = await this.prisma.quiz.findMany({
      where: {
        isCommunity: true,
        OR: [
          { name: { contains: search } },
          { description: { contains: search } },
          { difficulty: { name: { contains: search } } },
          { author: { username: { contains: search } } },
        ],
      },
      include: {
        category: true,
        difficulty: true,
        questions: {
          include: {
            answers: true,
          },
        },
        author: true,
      },
      skip,
      take,
    });

    // Retourner les quiz et le total
    return { quizzes, total };
  }

  async getCategories(): Promise<Category[]> {
    const categories = await this.prisma.category.findMany();
    return categories;
  }

  async saveCategory(name: string): Promise<void> {
    try {
      await this.prisma.category.upsert({
        where: { name },
        update: {},
        create: { name },
      });
    } catch (error) {
      throw new Error(`Failed to save category: ${error.message}`);
    }
  }

  async getInfo(quizId: string) {
    //get quiz info not questions or answers
    const quiz = await this.prisma.quiz.findUnique({
      where: { id: quizId },
      include: {
        category: true,
        difficulty: true,
      },
    });

    return quiz;
  }

  async getUserCustomQuiz(userId: string): Promise<Quiz[]> {
    const quizzes = await this.prisma.quiz.findMany({
      where: {
        authorId: Number(userId),
        isCommunity: true,
      },
      include: {
        category: true,
        difficulty: true,
        questions: {
          include: {
            answers: true,
          },
        },
      },
    });
    return quizzes;
  }
}
