import { Quiz, Run } from '@prisma/client';
import { RunRepository } from './run-repository';
import { PrismaService } from '../../../infra/prisma/prisma-service';
import { Injectable } from '@nestjs/common';
import { FullQuestion } from 'src/utils/FullQuestion';

@Injectable()
export class RunRepositoryOnDb implements RunRepository {
  constructor(private readonly prisma: PrismaService) {
  }

  async create(
    runID: string,
    quizId: string,
    originalQuizId: string,
    userId: number,
    partyId?: string,
  ): Promise<Run> {
    return this.prisma.run.create({
      data: {
        id: runID,
        quizId: quizId,
        originalQuizId: originalQuizId,
        quizUserId: userId as any,
        ...(partyId && { partyId }),
      },
    });
  }

  async getRunsByUser(userId: number) {
    const runs = await this.prisma.run.findMany({
      where: {
        quizUserId: userId,
      },
      include: {
        quiz: {
          select: {
            category: true,
            difficulty: true,
            name: true,
            questionCount: true,
            description: true,
          },
        },
      },
    });

    // Filtrer les `Run` avec `questionIndex < questionCount`
    return runs.filter((run) => run.questionIndex < run.quiz.questionCount);
  }



  async getQuestion(runId: string) {
    // Récupérer la `Run` pour obtenir l'index de la question actuelle et le quizId
    const run = await this.prisma.run.findUnique({
      where: { id: runId }, // Assurez-vous que l'ID est passé correctement ici
    });

    if (!run) {
      throw new Error(`Run with ID ${runId} not found`);
    }

    // Récupérer la question correspondante en utilisant l'index
    const question = await this.prisma.question.findFirst({
      where: {
        quizId: run.quizId, // Le quiz associé à la `Run`
        order: run.questionIndex - 1, // L'ordre correspond à l'index actuel
      },
      include: {
        answers: {
          select: {
            id: true,
            text: true, // Inclure uniquement les champs nécessaires
            questionId: true,
          },
        },
        category: true,
      },
    });

    if (!question) {
      throw new Error(`Vous avez atteint la fin du quiz avec l'ID ${runId}`);
    }

    return question;
  }

  async createWithoutLog(
    runID: string,
    quizId: string,
    originalQuizId: string,
    roomId?: string,
  ): Promise<any> {
    return this.prisma.run.create({
      data: {
        id: runID,
        quizId: quizId,
        originalQuizId: originalQuizId,
        quizUserId: null,
        ...(roomId && { roomId }),
      },
    });
  }

  async findRemainingQuestions(id: string): Promise<FullQuestion[]> {
    const questionIndex = (
      await this.prisma.run.findUnique({
        where: {
          id,
        },
        select: {
          questionIndex: true,
        },
      })
    ).questionIndex;

    const questions = await this.prisma.question.findMany({
      where: {
        quiz: {
          runs: {
            some: {
              id,
            },
          },
        },
      },
      include: {
        answers: true,
        category: true,
      },
      orderBy: {
        order: 'asc',
      },
    });

    // filter out questions that have already been answered
    return questions.filter((question) => question.order > questionIndex);
  }

  async findOne(id: string): Promise<Run | null> {
    return await this.prisma.run.findUnique({
      where: {
        id: id,
      },
    });
  }

  async incrementScore(id: string): Promise<void> {
    await this.prisma.run.update({
      where: {
        id,
      },
      data: {
        score: {
          increment: 1,
        },
      },
    });
  }

  async findById(id: string): Promise<any> {
    const run = await this.prisma.run.findUnique({
      where: {
        id,
      },
      include: {
        quiz: {
          include: {
            questions: {
              include: {
                answers: {
                  omit: {
                    isCorrect: true,
                  },
                },
                category: true,
              },
              orderBy: {
                order: 'asc',
              },
            },
          },
        },
      },
    });

    return run;
  }

  async incrementQuestionIndex(id: string): Promise<void> {
    await this.prisma.run.update({
      where: {
        id,
      },
      data: {
        questionIndex: {
          increment: 1,
        },
      },
    });
  }

  async getAnswers(id: string, questionID: number): Promise<FullQuestion> {
    return await this.prisma.question.findUnique({
      where: {
        id: questionID,
      },
      include: {
        answers: {
          omit: {
            isCorrect: false,
          },
        },
        category: true,
      },
    });
  }

  async getScore(id: string): Promise<number> {
    const run = await this.prisma.run.findUnique({
      where: {
        id,
      },
    });

    return run.score;
  }

  findAll(): Promise<Run[]> {
    return Promise.resolve([]);
  }

  async info(id: string): Promise<Run | null> {
    //récupérer les informations de la run avec l'id
    const run = await this.prisma.run.findUnique({
      where: {
        id,
      },
    });
    return run;
  }

  async getRunByHost(hostId: number): Promise<Run & { quiz: Quiz }> {
    const user = await this.prisma.quizUser.findUnique({
      where: {
        id: hostId,
      },
    });
    const run = await this.prisma.run.findFirst({
      where: {
        quizUserId: hostId,
        partyId: user.partyId,
      },
      include: {
        quiz: true,
      },
    });

    return run;
  }

  async findLastAnswered(runId: string): Promise<number> {
    const ra = await this.prisma.answerRun.findFirst({
      where: {
        runId,
      },
      orderBy: {
        questionId: 'desc',
      },
    });
    if (ra === null) {
      throw new Error('No question answered yet');
    }

    return ra.questionId;
  }
}
