import {
  BadRequestException,
  Injectable,
  InternalServerErrorException,
  NotFoundException,
} from '@nestjs/common';
import { UserRepository } from './user-repository';
import { QuizUser } from '@prisma/client';
import { PrismaService } from 'src/infra/prisma/prisma-service';
import { CreateUserDto } from '../dto/create-user.dto';
import { UpdateUserDto } from '../dto/update-user.dto';
import * as bcrypt from 'bcrypt';

type User = Omit<QuizUser, 'password'>;
@Injectable()
export class UserRepositoryOnDb implements UserRepository {
  constructor(private readonly prisma: PrismaService) {}

  async create(user: CreateUserDto): Promise<Omit<QuizUser, 'password'>> {
    try {
      // hash password using bcrypt

      const hashedPassword = await bcrypt.hash(user.password, 10);

      const newUser = await this.prisma.quizUser.create({
        data: {
          username: user.username,
          email: user.email,
          password: hashedPassword,
        },
        select: {
          id: true,
          username: true,
          email: true,
          partyId: true,
          isHost: true,
        },
      });

      return newUser;
    } catch (error) {
      if (error.code === 'P2002') {
        if (error.meta.target.includes('email')) {
          throw new BadRequestException('Email already taken');
        } else if (error.meta.target.includes('username')) {
          throw new BadRequestException('Username already taken');
        }
      } else {
        throw new InternalServerErrorException('Error creating user');
      }
    }
  }

  update(user: UpdateUserDto): Promise<User> {
    try {
      const updatedUser = this.prisma.quizUser.update({
        where: {
          id: user.id,
        },
        data: {
          username: user.username,
          email: user.email,
          password: user.password,
        },
      });

      return updatedUser;
    } catch {
      throw new InternalServerErrorException('Error updating user');
    }
  }

  async deleteById(id: number): Promise<void> {
    try {
      await this.prisma.quizUser.delete({
        where: {
          id,
        },
      });
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      throw new InternalServerErrorException('Error deleting user');
    }
  }

  async findById(id: number): Promise<User> {
    try {
      const user = this.prisma.quizUser.findUnique({
        where: {
          id,
        },
      });

      return user;
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      throw new InternalServerErrorException('Error finding user');
    }
  }

  async findByUsername(username: string): Promise<User> {
    try {
      const user = this.prisma.quizUser.findUnique({
        where: {
          username,
        },
      });

      return user;
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      throw new InternalServerErrorException('Error finding user');
    }
  }

  async findByLogin(login: string): Promise<User> {
    try {
      const user = this.prisma.quizUser.findFirst({
        where: {
          OR: [
            {
              username: login,
            },
            {
              email: login,
            },
          ],
        },
        omit: {
          password: false,
        },
      });

      return user;
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      throw new InternalServerErrorException('Error finding user');
    }
  }

  async findByEmail(email: string): Promise<User> {
    try {
      const user = this.prisma.quizUser.findUnique({
        where: {
          email,
        },
      });

      return user;
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      throw new InternalServerErrorException('Error finding user');
    }
  }

  async findManyByPartyId(partyId: string): Promise<User[]> {
    try {
      const users = await this.prisma.quizUser.findMany({
        where: {
          partyId: partyId,
        },
      });

      return users;
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      throw new InternalServerErrorException('Error finding user');
    }
  }

  async remove(id: number): Promise<User> {
    try {
      const user = await this.prisma.quizUser.delete({
        where: {
          id,
        },
      });

      return user;
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      throw new InternalServerErrorException('Error finding user');
    }
  }

  async getStats(id: number): Promise<any> {
    // get user stats: number of quizzes created, number of quizzes played, % of correct answers
    try {
      const stats = (await this.prisma.$queryRaw`
              SELECT
                (SELECT COUNT(*) FROM Quiz WHERE authorId = ${id}) as quizzesCreated,
                (SELECT COUNT(*) FROM QuizUser WHERE id = ${id}) as quizzesPlayed,
                (SELECT COUNT(*)
                FROM AnswerRun ar
                JOIN Answer a ON ar.answerId = a.id
                JOIN Run r ON ar.runId = r.id
                WHERE r.quizUserId = ${id}
                AND a.isCorrect = TRUE)
                as correctAnswers,
                (SELECT COUNT(*) FROM AnswerRun WHERE runId IN (SELECT id FROM Run WHERE quizUserId = ${id})) as totalAnswers
            `) as any;

      const formattedStats = {
        quizzesCreated: Number(stats[0].quizzesCreated),
        quizzesPlayed: Number(stats[0].quizzesPlayed),
        correctAnswers: Number(stats[0].correctAnswers),
        totalAnswers: Number(stats[0].totalAnswers),
      };

      return formattedStats;
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      console.error(error);
      throw new InternalServerErrorException('Error finding user');
    }
  }

  async findAll(): Promise<User[]> {
    try {
      const users = await this.prisma.quizUser.findMany();

      return users;
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      throw new InternalServerErrorException('Error finding user');
    }
  }

  async findOneWithPartyInfos(id: number) {
    try {
      const user = await this.prisma.quizUser.findUnique({
        where: {
          id,
        },
        include: {
          Party: true,
        },
      });

      return user;
    } catch (error) {
      if (error.code === 'P2025') {
        throw new NotFoundException('User not found');
      }
      throw new InternalServerErrorException('Error finding user');
    }
  }
}
